#! /usr/bin/env python3
#
# extract_replay_log.py 
#
# $Copyright(c) 2006 - 2008, 2011 Progress Software Corporation (PSC). All rights reserved. $
# $Copyright (c) 2013-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
# Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
#
# $Revision: 394340 $ $Date: 2021-08-23 19:26:09 +0100 (Mon, 23 Aug 2021) $
# 

import getopt, os.path, sys, xml.dom.minidom, binascii, base64, zipfile, io, re, subprocess
import ntpath, posixpath
def usage():
	print("Usage: extract_replay_log.py [options] <replayLogFile>")
	print()
	print("Unpack an input log so it can be replayed into a correlator.")
	print()
	print("  -o, --output=dir     Put output in directory dir, default is the current")
	print("                       directory")
	print("  -l, --lang=lang      Set the output language for the resulting script to")
	print("                       lang, default: batch. Choices are: shell, batch")
#	print("                       Also: pysys, apamainternal")
#	print("  -t, --testFramework  Generate python which can be run directly from within Apama internal framework")
#	print("  -p, --profile        Perform profiling around the last send operation")
#	print("  -s, --scenarios      Unpack scenario monitors and queries into the original source")
	print("  -c, --correlator     Start a correlator in the resulting script")
	print("  --licence            Specify a path to a licence file for starting a correlator")
	print("  --port               Specify a port to start the correlator on")
	print("  -v, --verbose        Produce verbose output")
	print("  -h, --help           Display this help message and exit")
	print()
	sys.exit();
	

################################################################################
#
# Internal methods and constants
#

REPLAY_ITEM_EVENT = 'EVNT'
REPLAY_ITEM_YAML = 'YAML'
REPLAY_ITEM_PROP = 'PROP'
REPLAY_ITEM_MONITOR = 'MONI'
REPLAY_ITEM_MONITOR_WITHFILENAME = 'MONF'
REPLAY_ITEM_JMON = 'JMON'
REPLAY_ITEM_CDP = 'CDPI'
REPLAY_ITEM_CDP_WITHFILENAME = 'CDPF'
REPLAY_ITEM_DELETE = 'DELE'
REPLAY_ITEM_FORCEDEL = 'FDEL'
REPLAY_ITEM_DELALL = 'DLAL'
REPLAY_ITEM_PDBN = 'PDBN'
REPLAY_ITEM_PDBJ = 'PDBJ'
REPLAY_ITEM_RTIM = 'RTIM'

REPLAY_ITEM_START_DELETE = 'SDEL'
REPLAY_ITEM_START_FORCEDEL = 'SFDL'
REPLAY_ITEM_START_DELALL = 'SDAL'

# Used internally by this tool - not read from replay log
REPLAY_ITEM_SCENARIO = 'SCENARIO'
REPLAY_ITEM_QUERY = 'QUERY'

def basename(s):
	b=os.path.basename(s)
	bs=b.split('\\')
	if len(bs)>1:
		b=bs[len(bs)-1]
	return b

def createFileName(itemType, filename=None):
	global replayFiles, outputFilePrefix, eventIdx
	eventIdx = eventIdx + 1
	name = outputFilePrefix+"%03d" % (eventIdx)
	suffix=''
	if itemType == REPLAY_ITEM_EVENT:
		name = name+".evt"
		suffix='.evt'
	elif itemType == REPLAY_ITEM_MONITOR:
		name = name+".mon"
		suffix='.mon'
	elif itemType == REPLAY_ITEM_MONITOR_WITHFILENAME:
		name = name+".mon"
		suffix='.mon'
	elif itemType == REPLAY_ITEM_PROP:
		name = name+".properties"
		suffix='.properties'
	elif itemType == REPLAY_ITEM_YAML:
		name = name+".yaml"
		suffix='.yaml'
	elif itemType == REPLAY_ITEM_JMON:
		name = name+".jar"
		suffix='.jar'
	elif itemType == REPLAY_ITEM_CDP:
		name = name+".cdp"
		suffix='.cdp'
	elif itemType == REPLAY_ITEM_CDP_WITHFILENAME:
		name = name+".cdp"
		suffix='.cdp'
	elif itemType == REPLAY_ITEM_PDBN:
		name = name+".db"
		suffix='.db'
	elif itemType == REPLAY_ITEM_PDBJ:
		name = name+".db-journal"
		suffix='.db-journal'		
	else:
		raise Exception("Unknown output file type: %s" % itemType)
	
	if filename:
		if language == "shell":
			name = posixpath.join(os.path.dirname(outputFilePrefix),basename(filename).replace(suffix, ".%03d%s" % (eventIdx, suffix)))
		else:
			name = os.path.join(os.path.dirname(outputFilePrefix),basename(filename).replace(suffix, ".%03d%s" % (eventIdx, suffix)))
		
	return name

def createReplayOutputFile(itemType, filename=None, addToReplayList=True):
	# Create a new .mon or .evt output replay output file 
	# Return a handle to the new file, which is always opened in binary mode
	name = createFileName(itemType, filename)
	if addToReplayList:
		replayFiles.append( (itemType, name, False) )
		verboseMessage("Writing output file %s ..." % name)
	return io.open(name, "wb+")

def closeEventFile(isInject=False):
	global eventFile, prevFile, eventLines
	if prevFile != None:
		prevFile.close()
		prevFile = None
	if eventFile != None:
		if isInject:
			prevFile=eventFile
			eventFile = None
			return
		# Will need to start a new event file next time we get an event
		eventFile.close() 
		eventFile = None
	eventLines = 0
	
def addReplayOutputName(itemType, name, filename):
	global replayFiles
	replayFiles.append( (itemType, name, False) )
	closeEventFile()
	
def hexToBin(hex):
	return binascii.unhexlify(hex) # returns byte string

def splitPath(path):
	s = os.path.split(path)
	if s[0] == '':
		return [s[1]]
	else:
		x = splitPath(s[0])
		x.append(s[1])
		return x

def getScenarioConfig(dirname, names):
	blockCatalogs = []
	functionCatalogs = []
	for name in names:
		if os.path.splitext(name)[1].lower() == '.sdf':
			scenario = name
			continue
		s = splitPath(name)
		if s[0] == 'blocks':
			if s[1] not in blockCatalogs:
				blockCatalogs.append(s[1])
		elif s[0] == 'functions':
			if s[1] not in functionCatalogs:
				functionCatalogs.append(s[1])
		else:
			raise Exception('Unexpected directory in scenario jar: ' + s[0])
	return (dirname, scenario, blockCatalogs, functionCatalogs)
	
def getJar(content, filename):
	lines = content.splitlines()
	for i in range(len(lines)):
		if lines[i].find('// SCENARIO SOURCE:') == 0:
			break
	if i == len(lines) - 1:
		return None
	b64 = io.StringIO()
	for j in range(i + 1, len(lines)):
		b64.write(lines[j][3:] + '\n')
	return zipfile.ZipFile(io.StringIO(base64.b64decode(b64.getvalue())))

def processScenario(type, content, filename, addToReplayList=True):
	jar = getJar(content, filename)
	if jar is None:
		return False
	dir = os.path.splitext(createFileName(REPLAY_ITEM_MONITOR, filename))[0]
	os.mkdir(dir)
	jar.extractall(dir)
	closeEventFile(True)
	if addToReplayList:
		replayFiles.append( (REPLAY_ITEM_SCENARIO, getScenarioConfig(dir, jar.namelist()), False) )
		return True
	return False 

def processQuery(type, content, filename, addToReplayList=True):
	lines = content.splitlines()
	for i in range(len(lines)):
		if lines[i].find('// QUERY SOURCE:') == 0:
			break
	if i == len(lines) - 2:
		return None
	b64 = io.StringIO()
	for j in range(i + 2, len(lines)):
		b64.write(lines[j][3:] + '\n')
	b64_decoded = base64.b64decode(b64.getvalue()).decode(encoding='UTF-8') # produce a character string
	filename = os.path.splitext(createFileName(REPLAY_ITEM_MONITOR, filename))[0] + ".qry"
	with io.open(filename, mode='w+', encoding='utf-8') as f:
		f.write(b64_decoded)
	if addToReplayList:
		replayFiles.append( (REPLAY_ITEM_QUERY, filename, False) )
		return True
	return False

def processMonitor(type, content, filename):
	if scenarios:
		if content.find('package Scenario_') != -1:
			addScenarioToReplayList = (language in ["apamainternal", "pysys"])
			if processScenario(type, content, filename, addScenarioToReplayList):
				# If scenario has been created and added to replayList, there's no need to write the MS to the replay script; the replay
				# script will regenerate it for us
				return
		if content.find('// QUERY SOURCE:') != -1:
			if processQuery(type, content, filename, addToReplayList = (language == "apamainternal")):
				# If the query has been created and added to replayList, there's no need to write the MS to the replay script; the replay
				# script will regenerate it for us
				return
			
	monfile = createReplayOutputFile(REPLAY_ITEM_MONITOR, filename)
	monfile.write(content.encode('utf-8'))
	monfile.close()
	closeEventFile(True)
	
def processYAML(type, content, filename):
	yamlFile = createReplayOutputFile(REPLAY_ITEM_YAML, filename)
	content = re.sub('(.*)storeLocation:.*', r'\1storeLocation: ${PARENT_DIR}', content)
	yamlFile.write(content.encode('utf-8'))
	yamlFile.close()
	closeEventFile(True)

def processProperties(type, content, filename):	
	propFile = createReplayOutputFile(REPLAY_ITEM_PROP, filename)
	propFile.write(content.encode('utf-8'))
	propFile.close()
	closeEventFile(True)

def processJMon(type, content, filename):
	jmonfile = createReplayOutputFile(REPLAY_ITEM_JMON)
	content = hexToBin(content)
	jmonfile.write(content)
	jmonfile.close()
	closeEventFile(True)
	
def processCDP(type, content, filename):
	cdpfile = createReplayOutputFile(REPLAY_ITEM_CDP)
	content = hexToBin(content)
	cdpfile.write(content)
	cdpfile.close()
	closeEventFile(True)
	
def processCDPWithFilename(type, content, filename):
	cdpfile = createReplayOutputFile(REPLAY_ITEM_CDP_WITHFILENAME, filename)
	content = hexToBin(content)
	cdpfile.write(content)
	cdpfile.close()
	closeEventFile(True)
	
# replace newlines in strings with \n and newlines outside of strings with nothing
def fixNewlines(event):
	olist = []

	instr = False
	escape = False

	for c in event:
		if '"' == c:
			olist.append(c)
			if instr and escape:
				escape = False
				continue
			instr = not instr
		elif escape:
			olist.append(c)
			escape = False
		elif '\\' == c:
			olist.append(c)
			escape = True
		elif '\n' == c:
			if instr:
				olist.append(r"\n")
		elif '\r' == c:
			if instr:
				olist.append(r"\r")
		else:
			olist.append(c)

	return ''.join(olist)

AP_MAX_REPLAY_LOG_EVENTS_PER_FILE = int(os.getenv('AP_MAX_REPLAY_LOG_EVENTS_PER_FILE','50000'))
def processActualEvent(event):
	global eventFile, lastSend, prevFile, eventLines
	if prevFile:
		if event.startswith('&REPLAY_ID'):
			# workaround for a bug in 4.1.0.0-4.1.0.1: the correlator may write inject
			prevFile.write(event.encode('utf-8'))
			prevFile.write('\n'.encode('utf-8'))
			return
		else:
			prevFile.close()
			prevFile = None
	
	# break up enormous blocks of events into reasonably-sized chunks
	if eventLines==AP_MAX_REPLAY_LOG_EVENTS_PER_FILE-1:
		closeEventFile()
	else:
		eventLines += 1
	
	if eventFile == None:
		eventFile = createReplayOutputFile(REPLAY_ITEM_EVENT)
		lastSend = len(replayFiles)-1
	
	# If there are newlines or carriage returns in the event, they must be escaped in order that the event stream reader will cope (it reads line-by-line).
	event = fixNewlines(event)

	eventFile.write(event.encode('utf-8'))
	eventFile.write('\n'.encode('utf-8'))
	
def processEvent(type, content, filename):
	global version

	if version == 1:
		event = content
	else:
		event = content.split(' ',1)[1]

	processActualEvent(event)
		
def processEnqueue(type, content, filename):
	# Assume enqueue has been turned into a NOP in the replay correlator, and 
	# just send it in like any other event
	processActualEvent(content)

def processTime(type, content, filename):
	global lastTime
	# Assume correlator is externally clocked:
	if type == 'STIM':
		processActualEvent('&SETTIME(%s)' % content)
	else:
		content=content.split(',')[0]
		t=float(content)
		if t <= lastTime:
			return
		else:
			lastTime = t
		processActualEvent('&TIME(%s)' % content)

def processRecoveryTime(type, content, filename):
	global recoveryTime, haveRecoveryTime
	print(("RecoveryTime:", content))
	recoveryTime = float(content)
	haveRecoveryTime = True

def processVersion(type, content, filename):
	global version
	print('Replay version: %s' % content)
	version = int(content)
	
def processHeader(type, content, filename):
	global javaEnabled, persistentCorrelator
	dom = xml.dom.minidom.parseString(content)
	print(dom.toxml())
	javaEnabled = dom.getElementsByTagName('javaEnabled')[0].childNodes[0].nodeValue == 'true'
	logMode = 'inputLog'
	try:
		logMode = dom.getElementsByTagName('replayLogMode')[0].childNodes[0].nodeValue
		if logMode == 'replayLog':
			raise Exception("ReplayLog mode no longer supported; use extract_replay_log from the same Apama version as generated the replay log.")
	except:
		pass
	print("Replay log mode assumed to be:", logMode)
		
	# it's useful to print the args and native/java path env vars when building the script
	print()
	print('Correlator arguments:')
	for node in dom.getElementsByTagName('args')[0].childNodes:
		if node.childNodes.length > 0:
			print('  ',node.childNodes[0].nodeValue)
			if (node.childNodes[0].nodeValue == '-P'):
				persistentCorrelator = True

	for node in dom.getElementsByTagName('environment')[0].childNodes:
		if node.childNodes.length > 0:
			x = node.childNodes[0].nodeValue
			if 'PATH=' in x: # classpath, path, ld_library_path
				print('Environment variable: '+x)
	print()

def processRandomSeed(type, content, filename):
	global seed, haveSeed
	print("RandomSeed:", content)
	seed = int(content)
	haveSeed = True


def processNow(dispatcher, type, content, filename):
	dispatcher(type, content, filename)

def processPDBN(type, content, filename):
	global haveDatabase, dbName
	dbName = createFileName(REPLAY_ITEM_PDBN)
	replayFiles.append ( (REPLAY_ITEM_PDBN, dbName, False) )
	verboseMessage("Writing output file %s ..." % dbName)
	pdbnfile = io.open(dbName, "wb+")
	content = hexToBin(content)
	pdbnfile.write(content)
	pdbnfile.close()
	closeEventFile(True)
	haveDatabase=True
	
def processPDBJ(type, content, filename):
	pdbjfile = createReplayOutputFile(REPLAY_ITEM_PDBJ)
	content = hexToBin(content)
	pdbjfile.write(content)
	pdbjfile.close()
	closeEventFile(True)
	
def filterAmpTime(dispatchtuple, type, content, filename):
	global version
	(subdispatcher, dispatcher)=dispatchtuple
	if version == 1:
		event = content
	else:
		event = content.split(' ',1)[1]
	if event.startswith('&TIME') or event.startswith('&SETTIME'):
		return
	subdispatcher(dispatcher, type, content, filename)

def ignoreMessage(type, content, filename):
	pass
	
def skipSpace(replayLog):
	ws = replayLog.read(1).decode('utf-8')
	if ws != ' ':
		raise Exception('Unexpected character: ' + ws)
		
def skipNewline(replayLog):
	ws = replayLog.read(1).decode('utf-8')
	if ws != '\n':
		raise Exception('Unexpected character: `' + ws +"' ("+str(ord(ws))+")")
		replayLog.readlines()
		for line in f:
			print(line, end=' ')


def buildDispatch(dispatch):
	queuedDispatch = {}
	notQueued = ['VERS', 'HEAD', 'CONN', 'DISC', 'TIME', 'STIM', REPLAY_ITEM_DELETE, REPLAY_ITEM_FORCEDEL,
		     REPLAY_ITEM_DELALL, REPLAY_ITEM_START_DELETE, REPLAY_ITEM_START_FORCEDEL, REPLAY_ITEM_START_DELALL]
	for key in list(dispatch.keys()):
		queuedDispatch[key]=(processNow, dispatch[key])
	queuedDispatch[REPLAY_ITEM_EVENT] = (filterAmpTime, queuedDispatch[REPLAY_ITEM_EVENT])
	return queuedDispatch
		

typeDispatch = {
	REPLAY_ITEM_EVENT : processEvent,
	REPLAY_ITEM_MONITOR : processMonitor,
	REPLAY_ITEM_MONITOR_WITHFILENAME : processMonitor,
	REPLAY_ITEM_PROP: processProperties,
	REPLAY_ITEM_YAML: processYAML,
	REPLAY_ITEM_JMON : processJMon,
	REPLAY_ITEM_CDP : processCDP,
	REPLAY_ITEM_CDP_WITHFILENAME : processCDPWithFilename,
	REPLAY_ITEM_PDBN : processPDBN,
	REPLAY_ITEM_PDBJ : processPDBJ,
	'TIME' : processTime,
	'STIM' : processTime,
	'ENQU' : processEnqueue,
	REPLAY_ITEM_DELETE : addReplayOutputName,
	REPLAY_ITEM_FORCEDEL : addReplayOutputName,
	REPLAY_ITEM_DELALL : addReplayOutputName,
	REPLAY_ITEM_START_DELETE: addReplayOutputName,
	REPLAY_ITEM_START_FORCEDEL: addReplayOutputName,
	REPLAY_ITEM_START_DELALL: addReplayOutputName,
	'VERS': processVersion,
	'HEAD': processHeader,
	'RAND': processRandomSeed,
	'CONN' : ignoreMessage,
	'DISC' : ignoreMessage,
	REPLAY_ITEM_RTIM : processRecoveryTime
}


javaEnabled = True
version = 1
persistentCorrelator = False

def addWaitingOperation(pendingOperations, type, name):
	n=(type,name)
	if n in pendingOperations:
		pendingOperations[n]=pendingOperations[n]+1
	else:
		pendingOperations[n]=1

def finishedWaitingOperation(pendingOperations, type, name):
	global version

	if version < 3:
		return
	n=(type,name)
	if n in pendingOperations:
		pendingOperations[n]=pendingOperations[n]-1
		if pendingOperations[n]==0:
			del pendingOperations[n]
	else:
		raise Exception("Operation without start")


def checkForPendingOperations(replayFiles):
	pendingOperations={}
	newReplayFiles=[]
	for (itemType, fileName, isLastSend) in replayFiles:
		add=True
		if itemType == REPLAY_ITEM_START_DELETE:
			addWaitingOperation(pendingOperations, REPLAY_ITEM_DELETE, fileName)
			add=False
		elif itemType == REPLAY_ITEM_START_FORCEDEL:
			addWaitingOperation(pendingOperations, REPLAY_ITEM_FORCEDEL, fileName)
			add=False
		elif itemType == REPLAY_ITEM_START_DELALL:
			addWaitingOperation(pendingOperations, REPLAY_ITEM_DELALL, "")
			add=False
		elif itemType == REPLAY_ITEM_DELETE:
			finishedWaitingOperation(pendingOperations, REPLAY_ITEM_DELETE, fileName)
		elif itemType == REPLAY_ITEM_FORCEDEL:
			finishedWaitingOperation(pendingOperations, REPLAY_ITEM_FORCEDEL, fileName)
		elif itemType == REPLAY_ITEM_DELALL:
			finishedWaitingOperation(pendingOperations, REPLAY_ITEM_DELALL, "")
		if add:
			newReplayFiles.append((itemType, fileName, isLastSend))
	for n in list(pendingOperations.keys()):
		print("Warning: pending operation %s (%i of)" % (n, pendingOperations[n]))
		for i in range(0, pendingOperations[n]):
			newReplayFiles.append((n[0], n[1], False))
	return newReplayFiles
		

################################################################################
#
# extractReplayLog method
#

def extractReplayLog(replayLogFileName, outputFileNamePrefix='./replay_'):
	""" Convert the specified replay log to a series of .evt and .mon files. 
	Returns an array of (itemType, outputFileName) tuples.
	
		replayLogFileName -- input file.
		outputFilePrefix -- directory/filename prefix for output replay files.
	"""
	global eventIdx, eventFile, replayFiles, outputFilePrefix, lastSend, prevFile, typeDispatch, eventLines
	# Initialize globals for this method invocation
	lastSend = -1
	eventIdx = 0
	eventFile = None
	prevFile = None
	eventLines = 0
	replayFiles = []
	outputFilePrefix = outputFileNamePrefix
	
	replayLogDirectory = os.path.dirname(replayLogFileName)
	
	replayLog = io.open(replayLogFileName, 'rb')
	
	try:
		typeDispatch = buildDispatch(typeDispatch)
		while 1: # Loop through entire input file
			
			# Read in type
			type = replayLog.read(4)
			if len(type) < 4:
				break
			type = type.decode('utf-8')
			skipSpace(replayLog)
			
			# Read in size of content
			lengths = replayLog.read(8)
			if len(lengths) != 8:
				raise Exception('Early EOF')
			lengths = lengths.decode('utf-8')
			skipSpace(replayLog)

			# Read in content
			length = int(lengths, 16)
			content = replayLog.read(length)
			if len(content) != length:
				raise Exception('Early EOF')
			content = content.decode('utf-8')	#Decode after length check. To Handle non-english characters
			filename = None
			
			if type in ['MONF'] or type in ['CDPF'] or type in ['PROP'] or type in ['YAML']:
				skipSpace(replayLog)
				filenamelengths = replayLog.read(8)
				if len(filenamelengths) != 8:
					raise Exception('Early EOF')
				filenamelengths = filenamelengths.decode('utf-8')
				skipSpace(replayLog)
				filenameLength = int(filenamelengths, 16)
				filename = replayLog.read(filenameLength)
				if len(filename) != filenameLength:
					raise Exception('Early EOF')
				filename = filename.decode('utf-8')
					
			skipNewline(replayLog)
			
			# Process content
			if type == 'NLOG':
				replayLogFileName = os.path.join(replayLogDirectory, os.path.basename(content))
				replayLog.close()
				replayLog = io.open(replayLogFileName, 'rb')
			else:
				processWhen, dispatcher=typeDispatch[type]
				processWhen(dispatcher, type, content, filename)
		
	finally:
		replayLog.close()
		if eventFile != None:
			eventFile.close()
		if prevFile != None:
			prevFile.close()
	if lastSend >= 0:
		replayFiles[lastSend]=(replayFiles[lastSend][0],replayFiles[lastSend][1],True)
	return replayFiles
	
	
class BaseHandler(object):
	# Transform the command line output directory into the one that the event files should
	# be written into. Returns what is passed in for most handler types
	def getOutputDirectory(self, outputdir):
		return outputdir

	def getConfigAndReplayFiles(self, allReplayFiles):
		"""
		Return (configfilenames, replayitems)
		where replayitems is: (itemType, fileName, isLastSend)
		"""
		return [c[1] for c in allReplayFiles if c[0] in [REPLAY_ITEM_PROP, REPLAY_ITEM_YAML]], [c for c in allReplayFiles if c[0] not in [REPLAY_ITEM_PROP, REPLAY_ITEM_YAML]]

class ShellHandler(BaseHandler):
	def processFiles(self, allReplayFiles):
		global correlator, argFlag
		shellScriptFileName = outputFileNamePrefix+'execute.sh'
		shellScriptFile = io.open(shellScriptFileName, "w") # open with platform default encoding
		shellScriptFile.write('#!/bin/sh\n')
		
		configFiles , replayFiles = [], []
		for item in allReplayFiles:
			if item[0] == REPLAY_ITEM_PROP or item[0] == REPLAY_ITEM_YAML:
				configFiles.append(item)
			else:
				replayFiles.append(item)
		
		
		if startCorrelator:
			shellScriptFile.write('%s %s -f correlator.log' % (correlator,argFlag))
			for (itemType, fileName, isLastSend) in configFiles:
				shellScriptFile.write(' --config %s' % fileName)
			shellScriptFile.write(' &\n')
			shellScriptFile.write('%s --waitFor 30\n' % engineManage)

		for (itemType, fileName, isLastSend) in replayFiles:
			if itemType == REPLAY_ITEM_EVENT:
				if profile and isLastSend:
					shellScriptFile.write('%s -r "cpuProfile on"\n' % engineManage)
				shellScriptFile.write('%s --utf8 "%s"\n' % (engineSend,fileName))
			elif itemType == REPLAY_ITEM_MONITOR or itemType == REPLAY_ITEM_MONITOR_WITHFILENAME:
				shellScriptFile.write('%s --utf8 "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_JMON:
				shellScriptFile.write('%s -j "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_CDP:
				shellScriptFile.write('%s -c "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_CDP_WITHFILENAME:
				shellScriptFile.write('%s -c "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_DELETE:
				shellScriptFile.write('%s "%s"\n' % (engineDelete,fileName))
			elif itemType == REPLAY_ITEM_FORCEDEL:
				shellScriptFile.write('%s -F "%s"\n' % (engineDelete,fileName))
			elif itemType == REPLAY_ITEM_DELALL:
				shellScriptFile.write('%s -a -y\n' % engineDelete)
			elif itemType == REPLAY_ITEM_PDBN or itemType == REPLAY_ITEM_PDBJ :
				verboseMessage("Dont do anything for Database files")
			else:
				raise Exception("Unexpected item type: %s" % itemType)

		if profile:
			shellScriptFile.write('# Wait for correlator queues to drain, then run engine_management -r "cpuProfile get"\n')
		if startCorrelator:
			shellScriptFile.write('engine_management -p %i -r flushAllQueues\n' % corrPort)
			shellScriptFile.write('engine_management -p %i --shutdown "Closing down after running replay log"\n' % corrPort)
		shellScriptFile.close()
		verboseMessage("Writing shell script %s ..." % shellScriptFileName)

class BatchHandler(BaseHandler):
	def processFiles(self, allReplayFiles):
		global correlator
		shellScriptFileName = outputFileNamePrefix+'execute.bat'
		shellScriptFile = io.open(shellScriptFileName, "w") # default encoding
		
		configFiles , replayFiles = [], []
		for item in allReplayFiles:
			if item[0] == REPLAY_ITEM_PROP or item[0] == REPLAY_ITEM_YAML:
				configFiles.append(item)
			else:
				replayFiles.append(item)
		
		if startCorrelator:
			shellScriptFile.write('start "correlator" /min %s %s -f correlator.log' % (correlator,argFlag))
			for (itemType, fileName, isLastSend) in configFiles:
				shellScriptFile.write(' --config %s' % fileName)
			shellScriptFile.write(' \n')
			shellScriptFile.write('%s --waitFor 30 \n' % engineManage)

		for (itemType, fileName, isLastSend) in replayFiles:
			if itemType == REPLAY_ITEM_EVENT:
				if profile and isLastSend:
					shellScriptFile.write('%s -r "cpuProfile on"\n' % engineManage)
				shellScriptFile.write('%s --utf8 "%s"\n' % (engineSend,fileName))
			elif itemType == REPLAY_ITEM_MONITOR or itemType == REPLAY_ITEM_MONITOR_WITHFILENAME:
				shellScriptFile.write('%s --utf8 "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_JMON:
				shellScriptFile.write('%s -j "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_CDP:
				shellScriptFile.write('%s -c "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_CDP_WITHFILENAME:
				shellScriptFile.write('%s -c "%s"\n' % (engineInject,fileName))
			elif itemType == REPLAY_ITEM_DELETE:
				shellScriptFile.write('%s "%s"\n' % (engineDelete,fileName))
			elif itemType == REPLAY_ITEM_FORCEDEL:
				shellScriptFile.write('%s -F "%s"\n' % (engineDelete,fileName))
			elif itemType == REPLAY_ITEM_DELALL:
				shellScriptFile.write('%s -a -y\n' % engineDelete)
			elif itemType == REPLAY_ITEM_PDBN or itemType == REPLAY_ITEM_PDBJ :
				verboseMessage("Dont do anything for Database files")				
			else:
				raise Exception("Unexpected item type: %s" % itemType)

		if profile:
			shellScriptFile.write('REM Wait for correlator queues to drain, then run engine_management -r "cpuProfile get"\n')
		if startCorrelator:
			shellScriptFile.write('engine_management -p %i -r flushAllQueues\n' % corrPort)
			shellScriptFile.write('engine_management -p %i --shutdown "Closing down after running replay log"\n' % corrPort)
		shellScriptFile.close()
		verboseMessage("Writing BATCH script %s ..." % shellScriptFileName)
	
class PysysHandler(BaseHandler):
	def getOutputDirectory(self, outputdir):
		if not os.path.exists(os.path.dirname(outputdir)) or os.path.exists(outputdir):
			raise Exception('Output directory must be set to the path where the pysys test should be created; this should not already exist but its parent must: %s'%outputdir)
		subprocess.check_call([sys.executable, os.environ['APAMA_HOME']+'/bin/pysys.py', 'make', outputdir], cwd=os.path.dirname(outputdir))
		self.testName = outputdir
		return outputdir # used to be outputdir/Input
		
	def processFiles(self, allReplayFiles):
		global version, argFlag
		pythonScriptFileName = os.path.join(self.testName, 'pysystest.py')
		
		with io.open(pythonScriptFileName, "r", encoding='utf-8') as pythonScriptFile:
			header = pythonScriptFile.read()
			assert 'class PySysTest' in header, header
			header = header[:header.find('class PySysTest')].rstrip()
			
		with io.open(pythonScriptFileName, "w", encoding='utf-8') as pythonScriptFile:
			pythonScriptFile.write(header)
			
			# default python encoding is ascii, but use utf8 so user can get it working just in case any non-ascii chars are present
			pythonScriptFile.write("\n\n# Generated by extract_replay_log.py\n\n")
			
			pythonScriptFile.write("class PySysTest(pysys.basetest.BaseTest):\n\n")
			pythonScriptFile.write("\tdef execute(self):\n")
			
			configFiles, replayFiles = self.getConfigAndReplayFiles(allReplayFiles)
			
			config = '['+', \n\t\t\t'.join(['self.input+"/%s"'%os.path.basename(c) for c in configFiles])+']'
			pythonScriptFile.write("\t\tcorrelator = self.apama.startCorrelator('correlator-replay', Xclock=True, java=True, config="+config+", arguments=[")
			for a in argFlag.split(' '):
				if a not in ['-j', '--java', '-Xclock']:
					if a.startswith('-'):
						pythonScriptFile.write('\n\t\t\t')
					pythonScriptFile.write(repr(a)+", ")
			pythonScriptFile.write('\n\t\t])\n')
				
			for (itemType, fileName, isLastSend) in replayFiles:
				if itemType == REPLAY_ITEM_EVENT:
					if profile and isLastSend:
						pythonScriptFile.write('\t\tcorrelator.manage(arguments=["-r", "cpuProfile on"])\n')
					pythonScriptFile.write('\t\tcorrelator.send("%s", utf8=True)\n' % os.path.basename(fileName))
				elif itemType == REPLAY_ITEM_MONITOR or itemType == REPLAY_ITEM_MONITOR_WITHFILENAME:
					pythonScriptFile.write('\t\tcorrelator.injectMonitorscript("%s", utf8=True)\n' % os.path.basename(fileName))
				elif itemType == REPLAY_ITEM_SCENARIO:
					# Not actually fileName in this case:
					scenarioData = fileName
					dirName = os.path.basename(scenarioData[0])
					rootDir = os.path.join('..', '..', 'Input', dirName)
					blockDir = []
					functDir = []
					if len(scenarioData[2]):
						for b in scenarioData[2]:
							bTuple = (b, os.path.join(rootDir, 'blocks', b))
							blockDir.append(bTuple)
					if len(scenarioData[3]):
						for f in scenarioData[3]:
							fTuple = (f, os.path.join(rootDir, 'functions', f))
							functDir.append(fTuple)
					pythonScriptFile.write('\t\tcorrelator.injectScenario(filename="%s", filedir="%s", blocks=%s, functions=%s)\n' % (scenarioData[1], rootDir, blockDir, functDir))
				elif itemType == REPLAY_ITEM_JMON:
					pythonScriptFile.write('\t\tcorrelator.injectJMON("%s", args="-j")\n' % os.path.basename(fileName))
				elif itemType == REPLAY_ITEM_CDP:
					pythonScriptFile.write('\t\tcorrelator.injectCDP("%s", args="-c")\n' % os.path.basename(fileName))
				elif itemType == REPLAY_ITEM_CDP_WITHFILENAME:
					pythonScriptFile.write('\t\tcorrelator.injectCDP("%s", args="-c")\n' % os.path.basename(fileName))
				elif itemType == REPLAY_ITEM_DELETE:
					pythonScriptFile.write('\t\tcorrelator.delete(["%s"])\n' % fileName)
				elif itemType == REPLAY_ITEM_FORCEDEL:
					pythonScriptFile.write('\t\tcorrelator.delete(["%s"], force=True)\n' % fileName)
				elif itemType == REPLAY_ITEM_DELALL:
					pythonScriptFile.write('\t\tcorrelator.delete(all=True)\n')
				elif itemType == REPLAY_ITEM_PDBN or itemType == REPLAY_ITEM_PDBJ :
					verboseMessage("Dont do anything for Database files")					
				else:
					raise Exception("Unexpected item type: %s" % itemType)
					
			pythonScriptFile.write("\t\tcorrelator.flush()\n")
			if profile:
				pythonScriptFile.write('\t\tcorrelator.manage(arguments=["-r", "cpuProfile get"], stdout=os.path.join(self.output, "profile.log"))\n')
			pythonScriptFile.write("\n\tdef validate(self):\n")
			
			pythonScriptFile.write("\t\tself.logFileContents('correlator-replay.log', includes=[' ERROR .*'])\n")
			pythonScriptFile.write("\t\tself.logFileContents('correlator-replay.log', includes=[' WARN .*'])\n")
			pythonScriptFile.write("\t\tself.assertGrep('correlator-replay.log', expr=' ERROR .*', contains=False)\n")
			pythonScriptFile.write("\t\tself.assertGrep('correlator-replay.log', expr=' WARN .*', contains=False)\n")
			
		verboseMessage("Writing pysys run script %s ..." % pythonScriptFileName)
	
class ApamaInternalHandler(BaseHandler):
	def processFiles(self, allReplayFiles):
		global version, testArgFlag, javaEnabled
		pythonScriptFileName = outputFileNamePrefix+'execute.py'
		# default python encoding is ascii, but use utf8 so user can get it working just in case any non-ascii chars are present
		pythonScriptFile = io.open(pythonScriptFileName, "w", encoding='utf-8')
		pythonScriptFile.write('# For use with the Apama test framework.\n#\n#\n\n')

		sourceDir = ""
		if testFramework:
			sourceDir = ", filedir=self.pathOutputFile()"
	
		configFiles , replayFiles = [], []
		for item in allReplayFiles:
			if item[0] == REPLAY_ITEM_PROP or item[0] == REPLAY_ITEM_YAML:
				configFiles.append(item)
			else:
				replayFiles.append(item)
		
		if not testFramework:
			javaFlag = str(javaEnabled)
			pythonScriptFile.write('\t\tabortOnInjectError=False\n')
			pythonScriptFile.write('\t\tfailOnInjectWarn=False\n')
			
			config = ''
			if (configFiles):
				config += ', config=['
				for (itemType, fileName, isLastSend) in configFiles:
					config += "'%s', " % os.path.basename(fileName)
				config += ']'
			
			pythonScriptFile.write("\n\t\tself.startCorrelator(Xclock=True, args='%s', abortOnError=True, java=%s %s)\n\n" % (testArgFlag, javaFlag, config))
		else:
			pythonScriptFile.write('abortOnInjectError=False\n')
			pythonScriptFile.write('failOnInjectWarn=False\n')
			print("Use correlator arguments: %s" % testArgFlag)

		for (itemType, fileName, isLastSend) in replayFiles:
			# use base names, so test can be located anywhere on disk
			if not testFramework:
				pythonScriptFile.write('\t\t')
			if itemType == REPLAY_ITEM_EVENT:
				if profile and isLastSend:
					pythonScriptFile.write('self.manage(argList=["-r","cpuProfile on"])\n')
					if not testFramework:
						pythonScriptFile.write('\t\t')
				pythonScriptFile.write('self.send(   ["%s"], utf8=True%s )\n' % (os.path.basename(fileName), sourceDir))
			elif itemType == REPLAY_ITEM_MONITOR or itemType == REPLAY_ITEM_MONITOR_WITHFILENAME:
				pythonScriptFile.write('self.inject( ["%s"], utf8=True%s, abortOnError=abortOnInjectError, failOnWarn=failOnInjectWarn )\n' % (os.path.basename(fileName), sourceDir))
			elif itemType == REPLAY_ITEM_SCENARIO:
				# Not actually fileName in this case:
				scenarioData = fileName
				dirName = os.path.basename(scenarioData[0])
				scenarioName = os.path.splitext(scenarioData[1])[0]
				if testFramework:
					rootDir = 'self.pathOutputFile("' + dirName + '")'
				else:
					rootDir = 'self.pathInputFile("' + dirName + '")'
				args = []
				if len(scenarioData[2]):
					args.append('"-XaddBlockPaths"')
					pathlist=[]
					for b in scenarioData[2]:
						pathlist.append(rootDir + ' + "/' + 'blocks' + '/' + b)
					args.append('," + '.join(pathlist)+'"')
				if len(scenarioData[3]):
					args.append('"-XaddFunctionPaths"')
					pathlist=[]
					for f in scenarioData[3]:
						pathlist.append(rootDir + ' + "/' + 'functions' + '/' + f)
					args.append('," + '.join(pathlist)+'"')
				pythonScriptFile.write('self.generateMonitor(sdfname="%s", sdfdir=%s, args=[%s], generateDebug=False)\n' % (scenarioData[1], rootDir, ','.join(args)))
				if not testFramework:
					pythonScriptFile.write('\t\t')
				pythonScriptFile.write('self.inject( ["%s.mon"], utf8=True, abortOnError=abortOnInjectError, filedir=self.pathOutputFile())\n' % scenarioName)
			elif itemType == REPLAY_ITEM_QUERY:
				pythonScriptFile.write('self.injectElasticQuery(queryFileList=["%s"], abortOnError=abortOnInjectError)\n' % os.path.basename(fileName))
			elif itemType == REPLAY_ITEM_JMON:
				pythonScriptFile.write('self.inject( ["%s"], args="-j"%s)\n' % (os.path.basename(fileName), sourceDir))
			elif itemType == REPLAY_ITEM_CDP:
				pythonScriptFile.write('self.inject( ["%s"], abortOnError=abortOnInjectError, failOnWarn=failOnInjectWarn , args="-c"%s)\n' % (os.path.basename(fileName), sourceDir))
			elif itemType == REPLAY_ITEM_CDP_WITHFILENAME:
				pythonScriptFile.write('self.inject( ["%s"], abortOnError=abortOnInjectError, failOnWarn=failOnInjectWarn , args="-c"%s)\n' % (os.path.basename(fileName), sourceDir))
			elif itemType == REPLAY_ITEM_DELETE:
				pythonScriptFile.write('self.delete( "%s" )\n' % fileName)
			elif itemType == REPLAY_ITEM_FORCEDEL:
				pythonScriptFile.write('self.delete( "%s", args = "-F" )\n' % fileName)
			elif itemType == REPLAY_ITEM_DELALL:
				pythonScriptFile.write('self.delete( all=1 )\n')
			elif itemType == REPLAY_ITEM_PDBN or itemType == REPLAY_ITEM_PDBJ :
				verboseMessage("Dont do anything for Database files")
				pythonScriptFile.write('\n')
				
			else:
				raise Exception("Unexpected item type: %s" % itemType)

		pythonScriptFile.write("\n")
		if not testFramework:
			pythonScriptFile.write('\t\t')
		if version>=3:
			pythonScriptFile.write("self.manage(argList=['-r','flushAllQueues'])\n")
		else:
			pythonScriptFile.write("self.waitForCompleteEvent()\n")
		if profile:
			if not testFramework:
				pythonScriptFile.write('\t\t')
			pythonScriptFile.write('self.manage(argList=["-r","cpuProfile get"])\n')
		pythonScriptFile.close()
		verboseMessage("Writing python script %s ..." % pythonScriptFileName)

def verboseMessage(message):
	global verbose
	if verbose:
		print(message)
	
outputFileNamePrefix = 'replay_'
outputDirectory = '.'
verbose = False
language = 'batch'
testFramework = False
profile = False
startCorrelator = False
corrPort = 15903
correlator = 'correlator'
engineSend = 'engine_send'
engineInject = 'engine_inject'
engineDelete = 'engine_delete'
engineManage = 'engine_management'
lastTime = 0
scenarios = False
handlers = {
	'shell': ShellHandler(),
	'batch': BatchHandler(),
	'pysys': PysysHandler(),
	'apamainternal': ApamaInternalHandler(),
}
argFlag=''
testArgFlag=''
haveSeed=False
seed=0
haveRecoveryTime=False
recoveryTime=0.0
haveDatabase=False
dbName=''

################################################################################
#
# Entry point
#

def main():

	global outputFileNamePrefix, outputDirectory, verbose, language, testFramework, javaEnabled, profile, scenarios, startCorrelator, corrPort, engineSend, engineInject, engineDelete, engineManage, correlator, argFlag, version, haveSeed, seed, haveRecoveryTime, haveDatabase, dbName, recoveryTime, testArgFlag, persistentCorrelator, classpath
	classpath = None

	try:
		opts, args = getopt.getopt(sys.argv[1:], "ho:l:tcvps", ["help", "output=", "lang=", "testFramework", "verbose", "profile", "scenarios","correlator","port=","licence=","classpath="])
	except getopt.GetoptError as err:
		print(err)
		print()
		usage()
	for o, a in opts:
		if o in ["-h", "--help"]:
			usage()
		if o in ["-o", "--output"]:
			outputDirectory = a
		if o in ["-v", "--verbose"]:
			verbose = True
		if o in ["-l", "--lang"]:
			language = a
		if o in ["-t", "--testFramework"]:
			testFramework = True
			pysys=False
		if o in ["-p", "--profile"]:
			profile = True
		if o in ["-s", "--scenarios"]:
			scenarios = True
		if o in ["-c", "--correlator"]:
			startCorrelator = True
		if o in ["--port"]:
			corrPort = int(a)
			correlator = '%s -p %i' % (correlator,corrPort)
			engineManage = '%s -p %i' % (engineManage,corrPort)
			engineDelete = '%s -p %i' % (engineDelete,corrPort)
			engineInject = '%s -p %i' % (engineInject,corrPort)
			engineSend = '%s -p %i' % (engineSend,corrPort)
		if o in ["--licence"]:
			correlator = '%s -l "%s"' % (correlator,a)
		if o in ["--classpath"]:
			classpath = a

	if len(args) != 1:
		print("Must specify one, and only one, replay log file name")
		print()
		usage()
	replayLogFileName = args[0]
	
	handler = handlers[language]
	outputDirectory = handler.getOutputDirectory(outputDirectory)

	if not os.access(outputDirectory, os.F_OK):
		os.makedirs(outputDirectory)
	
	if language == "shell":
		outputFileNamePrefix = posixpath.join(outputDirectory, outputFileNamePrefix)
	else:
		outputFileNamePrefix = os.path.join(outputDirectory, outputFileNamePrefix)
	
	# Create replay output files
	replayFiles = extractReplayLog(replayLogFileName=replayLogFileName, outputFileNamePrefix=outputFileNamePrefix)
	
	replayFiles = checkForPendingOperations(replayFiles)
	
	argFlag='-XignoreEnqueue -Xclock'
	if haveSeed:
		argFlag = argFlag +  ' -XsetRandomSeed %d' % (seed)
	if haveRecoveryTime:
		argFlag = argFlag +  ' -XrecoveryTime %f' % (recoveryTime)

	if persistentCorrelator and not haveDatabase:
		argFlag = argFlag +  ' -P '
	elif persistentCorrelator and haveDatabase:
		s = os.path.split(dbName)
		# Assumes db files are in the default location (i.e. output directory)
		argFlag = argFlag +  ' -P -PstoreName=%s' % (os.path.basename(dbName))
	testArgFlag=argFlag
	
	if javaEnabled:
		argFlag = argFlag + ' -j'
		if classpath:
			argFlag = argFlag + ' --javaopt -Djava.class.path="%s"' % classpath
	handler.processFiles(replayFiles)

	# Done
	verboseMessage("Done.")

if __name__ == "__main__":
	main()
		
